#!/usr/bin/env bash
###
 # Clears out files matching the "backup naming syntax" in the given directory.
 #
 # File clearing is done by the following:
 #  * Any backups from the last two weeks are saved.
 #  * After that, only one backup a week is saved, up through two months out.
 #  * After that, only one backup a month is saved, up until a year out.
 #  * After that, one backup a year is saved.
 #
 #
 # Required Args:
 #  * Directory to Clear
 #
 # Optional Flags:
 #  * Force - Force script to execute without confirmation of provided args.
 #      * Accepted Flags: f / force
 #
 #
 # https://git.brandon-rodriguez.com/scripts/bash/server_management
 # Version 1.0
 ##


# Define possible flag values for script.
possible_flags=("-f" "-force")


# Import utils script.
. $(dirname ${0})/../utils.sh


# Stop on error.
set -e


# Global Variables.
directory=""


###
 # Script entry point.
 ##
function main () {

    # Check for provided args.
    validate_args

    # Confirm args if force flag was not provided.
    if [[ ! ${flags[@]} =~ "f" && ! ${flags[@]} =~ "force" ]]
    then
        confirm_args
    fi

    clear_backups
}


###
 # Validate provided script args.
 ##
function validate_args () {

    # Check for minimum number of args.
    if [[ ${#args[@]} != 1 ]]
    then
        if [[ ${#args[@]} < 1 ]]
        then
            echo -e "${text_red}Missing \"directory\" arg. This is the folder to clear backups from.${text_reset}"
            exit 0
        else
            echo -e "${text_red}Script only takes \"directory\" arg. This is the folder to clear backups from.${text_reset}"
        fi
    fi

     # Check that location exists.
    get_absolute_path "${args[0]}"
    directory="${return_value}"
    if [[ ! -d ${directory} ]]
    then
        echo -e "${text_red}Provided location is not a folder, or does not exist.${text_reset}"
        echo -e "${text_red}Please verify the location exists, and that the location points to a folder, not a file.${text_reset}"
        exit 0
    fi
}


###
 # Prompts user to confirm input args.
 # Skipped if the -f flag is provided.
 ##
function confirm_args () {

    echo ""
    echo "Provided directory of \"${directory}\"."
    echo ""

    get_user_confirmation "Is this correct?"
    if [[ ${return_value} == false ]]
    then
        echo -e "${text_orange}Please correct provided args and rerun script.${text_reset}"
        exit 0
    fi
}


###
 # Logic to clear backups from folder.
 # Only clears files that match expected naming format.
 ##
function clear_backups () {

    local two_weeks_ago=$(date -d "$(date +%Y-%m-%d) -2 week" +%Y-%m-%d)
    local two_months_ago=$(date -d "$(date +%Y-%m-%d) -2 month" +%Y-%m-%d)
    local one_year_ago=$(date -d "$(date +%Y-%m-%d) -12 month" +%Y-%m-%d)

    # Loop through all entities in directory.
    for file in "${directory}/"*
    do
        # Check if file.
        if [[ -f ${file} ]]
        then
            # Parse to check if file matches expected naming format.
            parse_file_name ${file}
            local file_name=${return_value}
            local file_extension=${file_name#*.}
            local file_minus_extension=${file_name%%.*}
            local file_sub_name=${file_minus_extension%%__*}
            local file_date=${file_minus_extension##*__}

            # Verify file extension.
            if [[ ${file_extension} == ${file_name} ]]
            then
                # Use "fake" file extension "bkp".
                file_extension="bkp"
            fi

            # Check general naming format.
            if [[ "${file_sub_name}" != "" && "${file_date}" != "" && "${file_sub_name}" != "${file_date}" ]]
            then
                # Check date values.
                local file_date_year=${file_date:0:4}
                local file_date_month=${file_date:5:2}
                local file_date_day=${file_date:8:2}
                local valid_date=true
                if [[ ${file_date_year} =~ ^[0-9]+$ && ${file_date_month} =~ ^[0-9]+$ && ${file_date_day} =~ ^[0-9]+$ ]]
                then
                    # Valid general naming format and valid date values. Check if file should be cleared based on date.
                    # Note that yyyy-mm-dd format is automatically in both chronological and alphabetical format.
                    # Thus we can iterate through naturally, in ordering found via bash.
                    # This also means we can just compare date string values directly, without complicated date logic.

                    if [[ "${file_date}" < "${one_year_ago}" ]]
                    then
                        # File is more than a year old. Only one file per year is saved at this point.
                        local name_years_ago="${directory}/${file_sub_name}__${file_date_year}-01-01.${file_extension}"

                        # Check if another file already exists for this year.
                        if [[ -f "${name_years_ago}" ]]
                        then
                            # Verify current file is not file for year.
                            if [[ "${file}" != "${name_years_ago}" ]]
                            then
                                # File already exists for year. Clear current file.
                                rm ${file}
                            fi
                        else
                            # File does not exist for year. Rename current file accordingly.
                            mv ${file} "${name_years_ago}"
                        fi

                    elif [[ "${file_date}" < "${two_months_ago}" ]]
                    then
                        # File is more than two months old. Only one backup a month is saved at this point.
                        local name_months_ago="${directory}/${file_sub_name}__${file_date_year}-${file_date_month}-01.${file_extension}"

                        # Check if another file already exists for this month.
                        if [[ -f "${name_months_ago}" ]]
                        then
                            # Verify current file is not file for year.
                            if [[ "${file}" != "${name_months_ago}" ]]
                            then
                                # File already exists for month. Clear current file.
                                rm ${file}
                            fi
                        else
                            # File does not exist for month. Rename current file accordingly.
                            mv ${file} "${name_months_ago}"
                        fi

                    elif [[ "${file_date}" < "${two_weeks_ago}" ]]
                    then
                        # File is more than two weeks old. Only one backup a week is saved at this point.


                        # Get date of "most recent monday prior to date".
                        # Includes datetime itself possibly being a monday.
                        local formatted_date="${file_date_year}${file_date_month}${file_date_day}"
                        local day=$(date -d "${formatted_date}" +%u)
                        local last_monday=$(date -d "${formatted_date} -$(( day - 1)) days" +%Y-%m-%d)
                        local name_weeks_ago="${directory}/${file_sub_name}__${last_monday}.${file_extension}"

                        # Check if another file already exists for given monday.
                        if [[ -f "${name_weeks_ago}" ]]
                        then
                            # Verify current file is not file for year.
                            if [[ "${file}" != "${name_weeks_ago}" ]]
                            then
                                # File already exists for last monday. Clear current file.
                                rm ${file}
                            fi
                        else
                            # File does not exist for last monday. Rename current file accordingly.
                            mv ${file} "${name_weeks_ago}"
                        fi
                    fi
                fi

            fi
        fi
    done
}


main
